#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <stdint.h>
#include <fcntl.h>

#include "common.h"
#include "macros.h"
#include "misc.h"
#include "io.h"
#include "timeout.h"
#include "signal-event.h"
#include "buffer.h"
#include "hal-proto.h"
#include "hal-cmd-parser.h"
#include "hal-cmd.h"
#include "hal-peer.h"

#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
				__FUNCTION__, ##arg);

#define error(eval, fmt, arg...)			\
do {							\
	fprintf(stderr, "error: %s: "fmt CRLF,		\
				__FUNCTION__, ##arg);	\
	exit(eval);					\
} while (0)

static struct io_context *io;

static void connect_request_cb(void *opaque)
{
	struct hal_peer_base *p;

	p = (struct hal_peer_base *)opaque;

	info("get connect request");

	hal_cmd_connect_reply(p->wb, 32, 64, 0);
	hal_peer_send_default(p);
}

static void disconnect_request_cb(void *opaque)
{
	struct hal_peer_base *p;

	p = (struct hal_peer_base *)opaque;

	info("get disconnect request");

	hal_cmd_disconnect_reply(p->wb);
	hal_peer_send_default(p);
}

static void cycle_begin_request_cb(void *opaque)
{
	struct hal_peer_base *p;

	p = (struct hal_peer_base *)opaque;

	info("get cycle begin request");

	hal_cmd_cycle_begin_reply(p->wb);
	hal_peer_send_default(p);
}

static void cycle_run_request_cb(void *opaque)
{
	struct hal_peer_base *p;

	p = (struct hal_peer_base *)opaque;

	info("get cycle run request");

	hal_cmd_cycle_run_reply(p->wb);
	hal_peer_send_default(p);
}

static void cycle_end_request_cb(void *opaque)
{
	struct hal_peer_base *p;

	p = (struct hal_peer_base *)opaque;

	info("get cycle end request");

	hal_cmd_cycle_end_reply(p->wb);
	hal_peer_send_default(p);
}

static void init_handlers(struct hal_cmd_callbacks *callbacks)
{
	callbacks->connect_request_callback = connect_request_cb;
	callbacks->disconnect_request_callback = disconnect_request_cb;
	callbacks->cycle_begin_request_callback = cycle_begin_request_cb;
	callbacks->cycle_run_request_callback = cycle_run_request_cb;
	callbacks->cycle_end_request_callback = cycle_end_request_cb;	
}

static void handle_signals(int signo, siginfo_t *si, void *opaque)
{
	switch (signo) {
	case SIGINT:
	case SIGTERM:
		info("exiting by signal %d", signo);
		io_loop_stop(io);
		break;
	}
}

static void peer_event_handler(struct hal_peer_base *peer,
                               hal_peer_event_t event)
{
	switch (event) {
	case HAL_PEER_EVENT_ERROR:
	case HAL_PEER_EVENT_TIMEOUT:
		info("got error or timeout event");
		io_loop_stop(peer->io);
		break;
	case HAL_PEER_EVENT_READ:
	case HAL_PEER_EVENT_WRITE:
		break;
	}
}

int main(int argc, char **argv)
{
	struct hal_peer_base *p;
	struct hal_cmd_callbacks callbacks;
	int rfd, wfd;
	ret_t ret;

	if (argc < 2)
		error(1, "not enough arguments: write fifo read fifo");

	/* Open pipes and get file descriptors. */
	ret = pipe_open(argv[1], O_RDWR, &wfd);
	if (ret != ret_ok)
		error(1, "can't open write pipe");

	ret = pipe_open(argv[2], O_RDWR, &rfd);
	if (ret != ret_ok)
		error(1, "can't open read pipe");

	/* Set signal handlers. */
	sigevent_ignore(SIGPIPE);
	sigevent_add(SIGINT, handle_signals, NULL, 1);
	sigevent_add(SIGTERM, handle_signals, NULL, 1);
	sigevent_init();

	/* Init loop facilities. */
	timeout_init();
	io = io_new();

	/* Create a single peer. */
	p = hal_peer_base_new();
	hal_peer_base_init(p, io, wfd, rfd, peer_event_handler);

	init_handlers(&callbacks);
	hal_cmd_parser_register_callbacks(&callbacks);

	/* Wait and parse commands. */
	io_loop_run(io);

	/* Free resources. */
	hal_peer_base_deinit(p);
	hal_peer_base_free(&p);

	io_free(&io);
	timeout_deinit();
	sigevent_deinit();

	return 0;
}

